! WHIZARD 2.2.0_alpha-4 Nov 15 2013
! 
! Copyright (C) 1999-2013 by 
!     Wolfgang Kilian <kilian@physik.uni-siegen.de>
!     Thorsten Ohl <ohl@physik.uni-wuerzburg.de>
!     Juergen Reuter <juergen.reuter@desy.de>
!     
!     with contributions by 
!     Christian Speckner <cnspeckn@googlemail.com> 
!     and  Felix Braam, Sebastian Schmidt, Daniel Wiesler 
!
! WHIZARD is free software; you can redistribute it and/or modify it
! under the terms of the GNU General Public License as published by 
! the Free Software Foundation; either version 2, or (at your option)
! any later version.
!
! WHIZARD is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this program; if not, write to the Free Software
! Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! This file has been stripped of most comments.  For documentation, refer
! to the source 'whizard.nw'

module compilations

  use iso_varying_string, string_t => varying_string !NODEP!
  use diagnostics !NODEP!
  use os_interface
  use unit_tests
  use variables
  use models
  use process_libraries
  use prclib_stacks
  use rt_data
  use process_configurations

  implicit none
  private

  public :: compile_library
  public :: compilations_test

  type :: compilation_t
     type(string_t) :: libname
     type(process_library_t), pointer :: lib => null ()
     logical :: recompile_library = .false.
     logical :: make_executable = .false.
   contains
     procedure :: init => compilation_init
     procedure :: compile_and_load => compilation_compile_and_load
  end type compilation_t




contains

  subroutine compilation_init (comp, libname, stack, var_list)
    class(compilation_t), intent(out) :: comp
    type(string_t), intent(in) :: libname
    type(prclib_stack_t), intent(inout) :: stack
    type(var_list_t), intent(in) :: var_list
    comp%libname = libname
    comp%lib => stack%get_library_ptr (comp%libname)
    if (.not. associated (comp%lib)) then
       call msg_fatal ("Process library '" // char (comp%libname) &
            // "' has not been declared.")
    end if
    comp%recompile_library = &
         var_list_get_lval (var_list, var_str ("?recompile_library"))
  end subroutine compilation_init

  subroutine compilation_compile_and_load (comp, os_data, force)
    class(compilation_t), intent(inout) :: comp
    type(os_data_t), intent(in) :: os_data
    logical, intent(in) :: force
    if (associated (comp%lib)) then
       call msg_message ("Process library '" &
            // char (comp%libname) // "': compiling ...")
       call comp%lib%configure ()
       call comp%lib%compute_md5sum ()
       call comp%lib%write_makefile (os_data, force)
       if (force)  call comp%lib%clean (os_data, distclean = .false.)
       call comp%lib%write_driver (force)
       call comp%lib%update_status (os_data)
       call comp%lib%load (os_data)
       call msg_message ("Process library '" // char (comp%libname) &
            // "': ... success.")
    end if
  end subroutine compilation_compile_and_load

  subroutine compile_library (libname, global)
    type(string_t), intent(in) :: libname
    type(rt_data_t), intent(inout), target :: global
    type(compilation_t) :: comp
    logical :: force
    force = var_list_get_lval (global%var_list, var_str ("?rebuild_library"))
    call comp%init (libname, global%prclib_stack, global%var_list)
    call comp%compile_and_load (global%os_data, force)
  end subroutine compile_library


  subroutine compilations_test (u, results)
    integer, intent(in) :: u
    type(test_results_t), intent(inout) :: results
    call test (compilations_1, "compilations_1", &
         "intrinsic test processes", &
         u, results)
    call test (compilations_2, "compilations_2", &
         "external process (omega)", &
         u, results)
end subroutine compilations_test

  subroutine compilations_1 (u)
    integer, intent(in) :: u
    type(string_t) :: libname
    type(rt_data_t), target :: global
    
    write (u, "(A)")  "* Test output: compilations_1"
    write (u, "(A)")  "*   Purpose: configure and compile test process"
    write (u, "(A)")

    call syntax_model_file_init ()

    call global%global_init ()

    libname = "compilation_1"
    call prepare_test_library (global, libname, 1)

    call compile_library (libname, global)
    
    call global%write_libraries (u)

    call global%final ()
    call syntax_model_file_final ()
    
    write (u, "(A)")
    write (u, "(A)")  "* Test output end: compilations_1"
    
  end subroutine compilations_1
  
  subroutine compilations_2 (u)
    integer, intent(in) :: u
    type(string_t) :: libname
    type(rt_data_t), target :: global
    
    write (u, "(A)")  "* Test output: compilations_2"
    write (u, "(A)")  "*   Purpose: configure and compile test process"
    write (u, "(A)")

    call syntax_model_file_init ()

    call global%global_init ()
    call var_list_set_log (global%var_list, var_str ("?omega_openmp"), &
         .false., is_known = .true.)

    libname = "compilation_2"
    call prepare_test_library (global, libname, 2)

    call compile_library (libname, global)
    
    call global%write_libraries (u)

    call global%final ()
    call syntax_model_file_final ()
    
    write (u, "(A)")
    write (u, "(A)")  "* Test output end: compilations_2"
    
  end subroutine compilations_2
  

end module compilations
