! WHIZARD 2.2.0_alpha-4 Nov 15 2013
! 
! Copyright (C) 1999-2013 by 
!     Wolfgang Kilian <kilian@physik.uni-siegen.de>
!     Thorsten Ohl <ohl@physik.uni-wuerzburg.de>
!     Juergen Reuter <juergen.reuter@desy.de>
!     
!     with contributions by 
!     Christian Speckner <cnspeckn@googlemail.com> 
!     and  Felix Braam, Sebastian Schmidt, Daniel Wiesler 
!
! WHIZARD is free software; you can redistribute it and/or modify it
! under the terms of the GNU General Public License as published by 
! the Free Software Foundation; either version 2, or (at your option)
! any later version.
!
! WHIZARD is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this program; if not, write to the Free Software
! Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! This file has been stripped of most comments.  For documentation, refer
! to the source 'whizard.nw'

module sf_lumi_linker

  use kinds, only: default !NODEP!
  use iso_varying_string, string_t => varying_string !NODEP!
  use file_utils !NODEP!
  use diagnostics !NODEP!
  use tao_random_numbers !NODEP!
  use lorentz !NODEP!
  use models
  use flavors
  use colors
  use quantum_numbers
  use state_matrices
  use polarizations
  use interactions
  use sf_aux
  use sf_base

  implicit none
  private

  public :: lumi_linker_data_t 



  type, extends (sf_data_t) :: lumi_linker_data_t 
     private 
     type(model_t), pointer :: model => null () 
     type(flavor_t), dimension(2) :: flv_in 
     integer, dimension(2) :: pdg_in
     real(default) :: sqrts 
   contains  
       procedure :: init => lumi_linker_data_init 
       procedure :: write => lumi_linker_data_write 
       procedure :: get_n_par => lumi_linker_data_get_n_par
       procedure :: get_pdg_out => lumi_linker_data_get_pdg_out
       procedure :: allocate_sf_int => lumi_linker_data_allocate_sf_int     
  end type lumi_linker_data_t 
 
  !!! JRR: WK please check
  type, extends (sf_int_t) :: lumi_linker_t
     type(lumi_linker_data_t), pointer :: data => null ()
     real(default) :: x = 0
     real(default) :: q = 0
   contains
     procedure :: type_string => lumi_linker_type_string
     procedure :: write => lumi_linker_write
     procedure :: complete_kinematics => lumi_linker_complete_kinematics
     procedure :: inverse_kinematics => lumi_linker_inverse_kinematics
     procedure :: init => lumi_linker_init
     procedure :: apply => lumi_linker_apply
  end type lumi_linker_t 
  



contains

  subroutine lumi_linker_data_init &
       (data, model, flv, sqrts)
    class(lumi_linker_data_t), intent(out) :: data 
    type(model_t), intent(in), target :: model
    type(flavor_t), dimension(2), intent(in) :: flv
    real(default), intent(in) :: sqrts 
    data%model => model 
    data%flv_in = flv 
    data%pdg_in = flavor_get_pdg (data%flv_in)
    data%sqrts = sqrts
     select case (char (model_get_name (data%model)))
    case ("QCD","Test")
       call msg_fatal ("Lumi linker not available for model " &
            // char (model_get_name (data%model)) // ".")
    end select         
  end subroutine lumi_linker_data_init 
 
  subroutine lumi_linker_data_write (data, unit, verbose)
    class(lumi_linker_data_t), intent(in) :: data 
    integer, intent(in), optional :: unit 
    logical, intent(in), optional :: verbose    
    integer :: u
    u = output_unit (unit);  if (u < 0)  return 
    write (u, "(1x,A)") "Lumi linker data:" 
    write (u, "(3x,A,A,A,A)") "  prt_in  = ", &
         char (flavor_get_name (data%flv_in(1))), &
         ", ", char (flavor_get_name (data%flv_in(2)))
    write (u, "(1x,A,ES19.12)") "  sqrts = ", data%sqrts 
  end subroutine lumi_linker_data_write 
 
  function lumi_linker_data_get_n_par (data) result (n)
    class(lumi_linker_data_t), intent(in) :: data
    integer :: n
    n = 2
  end function lumi_linker_data_get_n_par
  
!!! JRR: WK please check
  function lumi_linker_data_get_pdg_out (data) result (pdg_out)
    class(lumi_linker_data_t), intent(in) :: data
    integer, dimension(:), allocatable :: pdg_out
    integer :: i, n
    n = 2
    allocate (pdg_out (n))
    do i = 1, n
       pdg_out(i) = data%pdg_in(i)
    end do
  end function lumi_linker_data_get_pdg_out
  
  subroutine lumi_linker_data_allocate_sf_int (data, sf_int)
    class(lumi_linker_data_t), intent(in) :: data
    class(sf_int_t), intent(inout), allocatable :: sf_int
    allocate (lumi_linker_t :: sf_int)
  end subroutine lumi_linker_data_allocate_sf_int
  
  function lumi_linker_type_string (object) result (string)
    class(lumi_linker_t), intent(in) :: object
    type(string_t) :: string
    if (associated (object%data)) then
       string = "Lumi linker: beamstrahlungs spectrum" 
    else
       string = "Lumi linker: [undefined]"
    end if
  end function lumi_linker_type_string
  
  subroutine lumi_linker_write (object, unit)
    !!! JRR: WK please check
    !!! Guess these variables do not exist for the lumi linker (?)
    class(lumi_linker_t), intent(in) :: object
    integer, intent(in), optional :: unit
    integer :: u
    u = output_unit (unit)
    if (associated (object%data)) then
       call object%data%write (u)
       if (object%status >= SF_DONE_KINEMATICS) then
          write (u, "(1x,A)")  "SF parameters:"
          write (u, "(3x,A,ES17.10)")  "x =", object%x
          if (object%status >= SF_FAILED_EVALUATION) then
             write (u, "(3x,A,ES17.10)")  "Q =", object%q
          end if
       end if
       call object%base_write (u)
    else
       write (u, "(1x,A)")  "Lumi linker data: [undefined]"
    end if
  end subroutine lumi_linker_write
    
  subroutine lumi_linker_complete_kinematics (sf_int, x, f, r, map)
    !!! JRR: WK please check
    !!! This cannot be correct, as the lumi linker structure function has 
    !!! twice the variables (2->4 instead of 1->2 splitting)
    class(lumi_linker_t), intent(inout) :: sf_int
    real(default), dimension(:), intent(out) :: x
    real(default), intent(out) :: f
    real(default), dimension(:), intent(in) :: r
    logical, intent(in) :: map
    real(default) :: xb1
    if (map) then
       call msg_fatal ("Lumi linker: map flag not supported")
    else
       x(1) = r(1)
       f = 1
    end if
    xb1 = 1 - x(1)
    call sf_int%split_momentum (x, xb1)
    select case (sf_int%status)
    case (SF_DONE_KINEMATICS)
       sf_int%x = x(1)
    case (SF_FAILED_KINEMATICS)
       sf_int%x = 0
       f = 0
    end select
  end subroutine lumi_linker_complete_kinematics

  subroutine lumi_linker_inverse_kinematics (sf_int, x, f, r, map, set_momenta)
    !!! JRR: WK please check
    !!! This cannot be correct, as the lumi linker structure function has 
    !!! twice the variables (2->4 instead of 1->2 splitting)
    class(lumi_linker_t), intent(inout) :: sf_int
    real(default), dimension(:), intent(in) :: x
    real(default), intent(out) :: f
    real(default), dimension(:), intent(out) :: r
    logical, intent(in) :: map
    logical, intent(in), optional :: set_momenta
    real(default) :: xb1
    logical :: set_mom
    set_mom = .false.;  if (present (set_momenta))  set_mom = set_momenta
    if (map) then
       call msg_fatal ("Lumi linker: map flag not supported")
    else
       r(1) = x(1)
       f = 1
    end if
    xb1 = 1 - x(1)
    if (set_mom) then
       call sf_int%split_momentum (x, xb1)
       select case (sf_int%status)
       case (SF_DONE_KINEMATICS)
          sf_int%x = x(1)
       case (SF_FAILED_KINEMATICS)
          sf_int%x = 0
          f = 0
       end select
    end if
  end subroutine lumi_linker_inverse_kinematics

  subroutine lumi_linker_init (sf_int, data) 
    class(lumi_linker_t), intent(out) :: sf_int 
    class(sf_data_t), intent(in), target :: data
    logical, dimension(6) :: mask_h
    type(quantum_numbers_mask_t), dimension(6) :: mask 
    integer, dimension(6) :: hel_lock 
    type(polarization_t) :: pol1, pol2
    type(quantum_numbers_t), dimension(1) :: qn_fc1, qn_hel1, qn_fc2, qn_hel2
    type(flavor_t) :: flv_photon
    type(quantum_numbers_t) :: qn_photon, qn1, qn2
    type(quantum_numbers_t), dimension(6) :: qn
    type(state_iterator_t) :: it_hel1, it_hel2
    hel_lock = 0
    mask_h = .false.
    select type (data)
    type is (lumi_linker_data_t)
       hel_lock(1) = 5;  hel_lock(5) = 1;  mask_h(3) = .true.
       hel_lock(2) = 6;  hel_lock(6) = 2;  mask_h(4) = .true.
       mask = new_quantum_numbers_mask (.false., .false., mask_h)
       call sf_int%base_init (mask, [0._default, 0._default], &
            [0._default, 0._default], [0._default, 0._default], &
            hel_lock = hel_lock)
       sf_int%data => data
       call flavor_init (flv_photon, PHOTON, data%model)
       call quantum_numbers_init (qn_photon, flv_photon)
       call polarization_init_generic (pol1, data%flv_in(1))
       call quantum_numbers_init (qn_fc1(1), flv = data%flv_in(1))
       call polarization_init_generic (pol2, data%flv_in(2))
       call quantum_numbers_init (qn_fc2(1), flv = data%flv_in(2))
       call state_iterator_init (it_hel1, pol1%state) 
       do while (state_iterator_is_valid (it_hel1)) 
          qn_hel1 = state_iterator_get_quantum_numbers (it_hel1)
          qn1 = qn_hel1(1) .merge. qn_fc1(1) 
          qn(1) = qn1
          qn(3) = qn_photon;  qn(5) = qn1
          call state_iterator_init (it_hel2, pol2%state) 
          do while (state_iterator_is_valid (it_hel2)) 
             qn_hel2 = state_iterator_get_quantum_numbers (it_hel2) 
             qn2 = qn_hel2(1) .merge. qn_fc2(1) 
             qn(2) = qn2
             qn(4) = qn_photon;  qn(6) = qn2
             call interaction_add_state (sf_int%interaction_t, qn)
             call state_iterator_advance (it_hel2) 
          end do
          call state_iterator_advance (it_hel1)
       end do
       call polarization_final (pol1)
       call polarization_final (pol2)
       call interaction_freeze (sf_int%interaction_t) 
       !!! JRR: WK please check
       !!! Is this correct if the photon is the radiating particle??
       call sf_int%set_incoming ([1,2])
       call sf_int%set_radiated ([3,4])
       call sf_int%set_outgoing ([5,6])
       sf_int%status = SF_INITIAL
    end select
  end subroutine lumi_linker_init

  subroutine lumi_linker_apply (sf_int, scale) !!!, no_map) 
  !!! subroutine lumi_linker_apply (sf_int, r, no_map) 
    !!! JRR: WK please check
    !!! this somehow is different from the PDFs, so the types    
    !!! do not match. Guess we need two types of strfun,
    !!! 1->2 and 2->4, nez-pas?
    class(lumi_linker_t), intent(inout) :: sf_int
    !!! Dummy
    real(default), intent(in) :: scale
    !!!real(default), dimension(2), intent(in) :: r 
    !!!logical, intent(in) :: no_map
    !!!type(vector4_t), dimension(2) :: k
    !!!type(splitting_data_t), dimension(2) :: sd
    !!!real(default), dimension(2) :: m_in, x, xb
    real(default) :: f
    !!!type(vector4_t), dimension(4) :: q
    associate (data => sf_int%data)
      !!! !!! JRR: WKK please check
      !!! !!! No clue what to do here :(((
      !!! k(1) = interaction_get_momentum (sf_int%interaction_t, 1) 
      !!! k(2) = interaction_get_momentum (sf_int%interaction_t, 2) 
      !!! m_in = data%m_in
      !!! sd = new_splitting_data (k, m_in**2, m_in**2, m_in) 
      !!! call strfun (f, x, xb, r, lumi_linker_data, no_map)
      !!! call splitting_set_t_bounds (sd, x, xb) 
      !!! call splitting_set_collinear (sd)
      !!! q((/1, 3/)) = split_momentum (k(1), sd(1))
      !!! q((/2, 4/)) = split_momentum (k(2), sd(2))
      !!! call interaction_set_momenta (sf_int%interaction_t, q, outgoing=.true.) 
    end associate
    call interaction_set_matrix_element &
        (sf_int%interaction_t, cmplx (f, kind=default)) 
    sf_int%status = SF_EVALUATED
  end subroutine lumi_linker_apply
 

end module sf_lumi_linker
